# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
from sklearn.cluster import AgglomerativeClustering
from tsam.representations import representations


def segmentation(
    normalizedTypicalPeriods,
    noSegments,
    timeStepsPerPeriod,
    representationMethod=None,
    representationDict=None,
    distributionPeriodWise=True,
):
    """
    Agglomerative clustering of adjacent time steps within a set of typical periods in order to further reduce the
    temporal resolution within typical periods and to further reduce complexity of input data.

    :param normalizedTypicalPeriods: MultiIndex DataFrame containing the typical periods as first index, the time steps
        within the periods as second index and the attributes as columns.
    :type normalizedTypicalPeriods: pandas DataFrame

    :param noSegments: Number of segments in which the typical periods should be subdivided - equivalent to the number of
        inner-period clusters.
    :type noSegments: integer

    :param timeStepsPerPeriod: Number of time steps per period
    :type timeStepsPerPeriod: integer

    :returns:     - **segmentedNormalizedTypicalPeriods** (pandas DataFrame) --  MultiIndex DataFrame similar to
                    normalizedTypicalPeriods but with segments instead of time steps. Moreover, two additional index
                    levels define the length of each segment and the time step index at which each segment starts.
                  - **predictedSegmentedNormalizedTypicalPeriods** (pandas DataFrame) -- MultiIndex DataFrame with the same
                    shape of normalizedTypicalPeriods, but with overwritten values derived from segmentation used for
                    prediction of the original periods and accuracy indicators.
    """
    # Initialize lists for predicted and segmented DataFrame
    segmentedNormalizedTypicalPeriodsList = []
    predictedSegmentedNormalizedTypicalPeriodsList = []
    # do for each typical period
    for i in normalizedTypicalPeriods.index.get_level_values(0).unique():
        # make numpy array with rows containing the segmenatation candidates (time steps)
        # and columns as dimensions of the
        segmentationCandidates = np.asarray(normalizedTypicalPeriods.loc[i, :])
        # produce adjacency matrix: Each time step is only connected to its preceding and succeeding one
        adjacencyMatrix = np.eye(timeStepsPerPeriod, k=1) + np.eye(
            timeStepsPerPeriod, k=-1
        )
        # execute clustering of adjacent time steps
        if noSegments == 1:
            clusterOrder = np.asarray([0] * len(segmentationCandidates))
        else:
            clustering = AgglomerativeClustering(
                n_clusters=noSegments, linkage="ward", connectivity=adjacencyMatrix
            )
            clusterOrder = clustering.fit_predict(segmentationCandidates)
        # determine the indices where the segments change and the number of time steps in each segment
        segNo, indices, segmentNoOccur = np.unique(
            clusterOrder, return_index=True, return_counts=True
        )
        clusterOrderUnique = [clusterOrder[index] for index in sorted(indices)]
        # determine the segments' values
        clusterCenters, clusterCenterIndices = representations(
            segmentationCandidates,
            clusterOrder,
            default="meanRepresentation",
            representationMethod=representationMethod,
            representationDict=representationDict,
            distributionPeriodWise=distributionPeriodWise,
            timeStepsPerPeriod=1,
        )
        # clusterCenters = meanRepresentation(segmentationCandidates, clusterOrder)
        # predict each time step of the period by representing it with the corresponding segment's values
        predictedSegmentedNormalizedTypicalPeriods = (
            pd.DataFrame(clusterCenters, columns=normalizedTypicalPeriods.columns)
            .reindex(clusterOrder)
            .reset_index(drop=True)
        )
        # represent the period by the segments in the right order only instead of each time step
        segmentedNormalizedTypicalPeriods = (
            pd.DataFrame(clusterCenters, columns=normalizedTypicalPeriods.columns)
            .reindex(clusterOrderUnique)
            .set_index(np.sort(indices))
        )
        # keep additional information on the lengths of the segments in the right order
        segmentDuration = (
            pd.DataFrame(segmentNoOccur, columns=["Segment Duration"])
            .reindex(clusterOrderUnique)
            .set_index(np.sort(indices))
        )
        # create DataFrame with reduced number of segments together with three indices per period:
        # 1. The segment number
        # 2. The segment duration
        # 3. The index of the original time step, at which the segment starts
        result = segmentedNormalizedTypicalPeriods.set_index(
            [
                pd.Index(segNo, name="Segment Step"),
                segmentDuration["Segment Duration"],
                pd.Index(np.sort(indices), name="Original Start Step"),
            ]
        )
        # append predicted and segmented DataFrame to list to create a big DataFrame for all periods
        predictedSegmentedNormalizedTypicalPeriodsList.append(
            predictedSegmentedNormalizedTypicalPeriods
        )
        segmentedNormalizedTypicalPeriodsList.append(result)
    # create a big DataFrame for all periods for predicted segmented time steps and segments and return
    predictedSegmentedNormalizedTypicalPeriods = pd.concat(
        predictedSegmentedNormalizedTypicalPeriodsList,
        keys=normalizedTypicalPeriods.index.get_level_values(0).unique(),
    ).rename_axis(["", "TimeStep"])
    segmentedNormalizedTypicalPeriods = pd.concat(
        segmentedNormalizedTypicalPeriodsList,
        keys=normalizedTypicalPeriods.index.get_level_values(0).unique(),
    )
    return segmentedNormalizedTypicalPeriods, predictedSegmentedNormalizedTypicalPeriods



import os
import time

import pandas as pd
import numpy as np

import tsam.timeseriesaggregation as tsam


def test_hierarchical():

    raw = pd.read_csv(
        os.path.join(os.path.dirname(__file__), "..", "examples", "testdata.csv"),
        index_col=0,
    )

    orig_raw = pd.read_csv(
        os.path.join(
            os.path.dirname(__file__),
            "..",
            "examples",
            "results",
            "testperiods_hierarchical.csv",
        ),
        index_col=[0, 1],
    )

    starttime = time.time()

    aggregation = tsam.TimeSeriesAggregation(
        raw,
        noTypicalPeriods=8,
        hoursPerPeriod=24,
        clusterMethod="hierarchical",
        extremePeriodMethod="new_cluster_center",
        addPeakMin=["T"],
        addPeakMax=["Load"],
    )

    typPeriods = aggregation.createTypicalPeriods()

    print("Clustering took " + str(time.time() - starttime))

    # sort the typical days in order to avoid error assertion due to different order
    sortedDaysOrig = orig_raw.groupby(level=0).sum().sort_values("GHI").index
    sortedDaysTest = typPeriods.groupby(level=0).sum().sort_values("GHI").index

    # rearange their order
    orig = orig_raw[typPeriods.columns].unstack().loc[sortedDaysOrig, :].stack(future_stack=True,)
    test = typPeriods.unstack().loc[sortedDaysTest, :].stack(future_stack=True,)

    np.testing.assert_array_almost_equal(orig.values, test.values, decimal=4)

def test_hierarchical_for_weeks():

    raw = pd.read_csv(
        os.path.join(os.path.dirname(__file__), "..", "examples", "testdata.csv"),
        index_col=0,
    )

    starttime = time.time()

    aggregation = tsam.TimeSeriesAggregation(
        raw,
        noTypicalPeriods=8,
        hoursPerPeriod=24*7,
        clusterMethod="hierarchical",
        extremePeriodMethod="new_cluster_center",
        addPeakMin=["T"],
        addPeakMax=["Load"],
    )

    typPeriods = aggregation.createTypicalPeriods()

    print("Clustering took " + str(time.time() - starttime))

if __name__ == "__main__":
    test_hierarchical()



import os
import time

import pandas as pd
import numpy as np

import tsam.timeseriesaggregation as tsam


def test_samemean():

    raw = pd.read_csv(
        os.path.join(os.path.dirname(__file__), "..", "examples", "testdata.csv"),
        index_col=0,
    )
    # get all columns as floats to avoid warning
    for col in raw.columns:
        raw[col] = raw[col].astype(float)

    starttime = time.time()

    aggregation = tsam.TimeSeriesAggregation(
        raw,
        noTypicalPeriods=8,
        hoursPerPeriod=24,
        clusterMethod="k_means",
        sameMean=True,
    )

    typPeriods = aggregation.createTypicalPeriods()
    print("Clustering took " + str(time.time() - starttime))

    # test if the normalized time series all have the same mean
    means = aggregation.normalizedTimeSeries.mean().values
    np.testing.assert_allclose(means, np.array([means[0]] * len(means)), rtol=1e-5)

    # repredict the original data
    rearangedData = aggregation.predictOriginalData()

    # test if the mean fits the mean of the raw time series --> should always hold for k-means independent from sameMean True or False
    np.testing.assert_array_almost_equal(
        raw.mean(), rearangedData[raw.columns].mean(), decimal=4
    )






if __name__ == "__main__":
    test_samemean()



import os
import time

import pandas as pd
import numpy as np

import tsam.timeseriesaggregation as tsam


def test_accuracyIndicators():

    hoursPerPeriod = 24

    noTypicalPeriods = 8

    raw = pd.read_csv(
        os.path.join(os.path.dirname(__file__), "..", "examples", "testdata.csv"),
        index_col=0,
    )

    aggregation1 = tsam.TimeSeriesAggregation(
        raw,
        noTypicalPeriods=noTypicalPeriods,
        hoursPerPeriod=hoursPerPeriod,
        clusterMethod="hierarchical",
    )

    aggregation2 = tsam.TimeSeriesAggregation(
        raw,
        noTypicalPeriods=noTypicalPeriods,
        hoursPerPeriod=hoursPerPeriod,
        clusterMethod="hierarchical",
        sortValues=True,
    )

    # make sure that the sum of the attribute specific RMSEs is smaller for the normal time series clustering than for
    # the duration curve clustering
    np.testing.assert_array_less(
        aggregation1.accuracyIndicators().loc[:, "RMSE"].sum(),
        aggregation2.accuracyIndicators().loc[:, "RMSE"].sum(),
    )

    # make sure that the sum of the attribute specific duration curve RMSEs is smaller for the duration curve
    # clustering than for the normal time series clustering
    np.testing.assert_array_less(
        aggregation2.accuracyIndicators().loc[:, "RMSE_duration"].sum(),
        aggregation1.accuracyIndicators().loc[:, "RMSE_duration"].sum(),
    )

    # make sure that the same accounts for the total accuracy indicator
    np.testing.assert_array_less(
        aggregation1.totalAccuracyIndicators()["RMSE"],
        aggregation2.totalAccuracyIndicators()["RMSE"],
    )
    # make sure that the same accounts for the total accuracy indicator
    np.testing.assert_array_less(
        aggregation2.totalAccuracyIndicators()["RMSE_duration"],
        aggregation1.totalAccuracyIndicators()["RMSE_duration"],
    )


if __name__ == "__main__":
    test_accuracyIndicators()



import os
import time

import pandas as pd
import numpy as np

import tsam.timeseriesaggregation as tsam


def test_segmentation():

    raw = pd.read_csv(
        os.path.join(os.path.dirname(__file__), "..", "examples", "testdata.csv"),
        index_col=0,
    )

    orig_raw = pd.read_csv(
        os.path.join(
            os.path.dirname(__file__),
            "..",
            "examples",
            "results",
            "testperiods_segmentation.csv",
        ),
        index_col=[0, 1, 2],
    )

    starttime = time.time()

    aggregation = tsam.TimeSeriesAggregation(
        raw,
        noTypicalPeriods=20,
        hoursPerPeriod=24,
        clusterMethod="hierarchical",
        representationMethod="meanRepresentation",
        segmentation=True,
        noSegments=12,
    )

    typPeriods = aggregation.createTypicalPeriods()

    print("Clustering took " + str(time.time() - starttime))

    # sort the typical days in order to avoid error assertion due to different order
    sortedDaysOrig = orig_raw.groupby(level=0).sum().sort_values("GHI").index
    sortedDaysTest = typPeriods.groupby(level=0).sum().sort_values("GHI").index

    # rearange their order
    orig = orig_raw[typPeriods.columns].unstack().loc[sortedDaysOrig, :].stack(future_stack=True,)
    test = typPeriods.unstack().loc[sortedDaysTest, :].stack(future_stack=True,)

    np.testing.assert_array_almost_equal(orig.values, test.values, decimal=4)




def test_representation_in_segmentation():

    segmentationCandidates = np.array([[0.        , 0.38936961, 0.27539063, 0.25      ],
        [0.        , 0.35591778, 0.26841518, 0.25      ],
        [0.        , 0.35045773, 0.265625  , 0.25      ],
        [0.        , 0.36418749, 0.25372024, 0.25      ],
        [0.        , 0.38386857, 0.25167411, 0.25      ],
        [0.        , 0.42710529, 0.24237351, 0.16666667],
        [0.        , 0.5798638 , 0.23707217, 0.1922619 ],
        [0.        , 0.70166596, 0.24507068, 0.16666667],
        [0.01838546, 0.74661296, 0.24739583, 0.18363095],
        [0.06893491, 0.75398663, 0.26041667, 0.16666667],
        [0.0942519 , 0.77160913, 0.28385417, 0.16666667],
        [0.14374472, 0.80191153, 0.3046875 , 0.25      ],
        [0.11999155, 0.79502066, 0.3125    , 0.22678571],
        [0.10016906, 0.77613611, 0.31845238, 0.16666667],
        [0.07438715, 0.76489634, 0.3203125 , 0.16666667],
        [0.0101437 , 0.75082659, 0.31538318, 0.16666667],
        [0.        , 0.74856422, 0.3077567 , 0.16666667],
        [0.        , 0.76062049, 0.29678199, 0.08333333],
        [0.        , 0.78148316, 0.29427083, 0.16666667],
        [0.        , 0.75668439, 0.28738839, 0.16666667],
        [0.        , 0.67461737, 0.2859933 , 0.16666667],
        [0.        , 0.624061  , 0.28041295, 0.16666667],
        [0.        , 0.56076035, 0.2734375 , 0.16666667],
        [0.        , 0.4734255 , 0.27092634, 0.16666667]])

    clusterOrder = np.array([5, 5, 5, 5, 5, 7, 3, 2, 2, 2, 2, 0, 0, 0, 0, 1, 1, 1, 1, 1, 6, 6,
       4, 4])

    clusterCenters_mean, clusterCenterIndices = tsam.representations(
        segmentationCandidates,
        clusterOrder,
        default="meanRepresentation",
        representationMethod="meanRepresentation",
        distributionPeriodWise=False,
        timeStepsPerPeriod=1,
    )

    clusterCenters_dist, clusterCenterIndices = tsam.representations(
        segmentationCandidates,
        clusterOrder,
        default="meanRepresentation",
        representationMethod="distributionRepresentation",
        distributionPeriodWise=True,
        timeStepsPerPeriod=1,
    )

    assert np.isclose(clusterCenters_mean, clusterCenters_dist).all()


if __name__ == "__main__":
    test_segmentation()



import os
import time
import pytest

import pandas as pd
import numpy as np

import tsam.timeseriesaggregation as tsam
import tsam.hyperparametertuning as tune

def test_getPeriodPair():
    """Tests if the number of periods is properly defined if a datareduction is set
    """
    noRawTimeSteps=100
    segmentsPerPeriod=10
    dataReduction=0.5
    noPeriods=tune.getNoPeriodsForDataReduction(noRawTimeSteps, segmentsPerPeriod, dataReduction)
    assert noPeriods==5

    noRawTimeSteps=101
    noPeriods=tune.getNoPeriodsForDataReduction(noRawTimeSteps, segmentsPerPeriod, dataReduction)
    assert noPeriods==5
    
    segmentsPerPeriod=2
    noPeriods=tune.getNoPeriodsForDataReduction(noRawTimeSteps, segmentsPerPeriod, dataReduction)
    assert noPeriods==25
    
def test_optimalPair():

    raw = pd.read_csv(
        os.path.join(os.path.dirname(__file__), "..", "examples", "testdata.csv"),
        index_col=0,
    )

    datareduction=0.01

    # just take wind
    aggregation_wind = tune.HyperTunedAggregations(
        tsam.TimeSeriesAggregation(
            raw.loc[:, ['Wind']],
            hoursPerPeriod=24,
            clusterMethod="hierarchical",
            representationMethod="durationRepresentation",
            distributionPeriodWise=False,
            rescaleClusterPeriods=False,
            segmentation=True,
        )
    )

    # and identify the best combination for a data reduction of to ~10%. 
    windSegments, windPeriods, windRMSE= aggregation_wind.identifyOptimalSegmentPeriodCombination(dataReduction=datareduction)

    # just take solar irradiation
    aggregation_solar = tune.HyperTunedAggregations(
        tsam.TimeSeriesAggregation(
            raw.loc[:, ['GHI']],
            hoursPerPeriod=24,
            clusterMethod="hierarchical",
            representationMethod="durationRepresentation",
            distributionPeriodWise=False,
            rescaleClusterPeriods=False,
            segmentation=True,
        )
    )

    # and identify the best combination for a data reduction of to ~10%. 
    solarSegments, solarPeriods, solarRMSE = aggregation_solar.identifyOptimalSegmentPeriodCombination(dataReduction=datareduction)


    # according to Hoffmann et al. 2022 is for solar more segments and less days better than for wind
    assert windPeriods > solarPeriods
    assert windSegments < solarSegments

    # check if the number time steps is in the targeted range
    assert windPeriods * windSegments <= len(raw["Wind"])*datareduction
    assert windPeriods * windSegments >= len(raw["Wind"])*datareduction * 0.8

@pytest.mark.skip(reason="This test is too slow")
def test_steepest_gradient_leads_to_optima():
    """
    Based on the hint of Eva Simarik, check if the RMSE is for the optimized combination
    of segments and periods smaller than sole segmentation approach 
    """

    raw = pd.read_csv(
        os.path.join(os.path.dirname(__file__), "..", "examples", "testdata.csv"),
        index_col=0,
    )

    SEGMENTS_TESTED = 5

    datareduction = (SEGMENTS_TESTED*365)/8760

    # just take wind
    tunedAggregations = tune.HyperTunedAggregations(
        tsam.TimeSeriesAggregation(
            raw,
            hoursPerPeriod=24,
            clusterMethod="hierarchical",
            representationMethod="meanRepresentation",
            rescaleClusterPeriods=False,
            segmentation=True,
        )
    )

    # and identify the best combination for a data reduction. 
    segmentsOpt, periodsOpt, RMSEOpt = tunedAggregations.identifyOptimalSegmentPeriodCombination(dataReduction=datareduction)

    # test steepest
    tunedAggregations.identifyParetoOptimalAggregation(untilTotalTimeSteps=365*SEGMENTS_TESTED)
    steepestAggregation = tunedAggregations.aggregationHistory[-1]
    RMSEsteepest = steepestAggregation.totalAccuracyIndicators()["RMSE"]

    # only segments
    aggregation = tsam.TimeSeriesAggregation(raw,
        noTypicalPeriods= 365,
        hoursPerPeriod = 24,
        segmentation = True,
        noSegments = SEGMENTS_TESTED,
        clusterMethod="hierarchical",
        representationMethod="meanRepresentation",
    )
    
    RMSESegments = aggregation.totalAccuracyIndicators()["RMSE"]

    assert RMSEsteepest < RMSESegments

    assert np.isclose(RMSEsteepest, RMSEOpt, atol=1e-3)

    




def test_paretoOptimalAggregation():

    raw = pd.read_csv(
        os.path.join(os.path.dirname(__file__), "..", "examples", "testdata.csv"),
        index_col=0,
    )

    # reduce the set, since it takes otherwise too long
    raw=raw.iloc[:240,:]

    # set tuned aggregation
    tunedAggregations = tune.HyperTunedAggregations(
        tsam.TimeSeriesAggregation(
            raw,
            hoursPerPeriod=12,
            clusterMethod="hierarchical",
            representationMethod="meanRepresentation",
            distributionPeriodWise=False,
            rescaleClusterPeriods=False,
            segmentation=True,
        )
    )

    # determine pareto optimal aggregation
    tunedAggregations.identifyParetoOptimalAggregation()

    # test if last RMSE is 0
    assert tunedAggregations._RMSEHistory[-1] == 0

    # test if RMSE is continously decreasing
    for i, RMSE in enumerate(tunedAggregations._RMSEHistory[1:]):
        assert RMSE <= tunedAggregations._RMSEHistory[i]


if __name__ == "__main__":
    test_paretoOptimalAggregation()



import os
import time

import pandas as pd
import numpy as np

import tsam.timeseriesaggregation as tsam


def test_preprocess():

    raw = pd.read_csv(
        os.path.join(os.path.dirname(__file__), "..", "examples", "testdata.csv"),
        index_col=0,
    )

    raw_wind = raw.loc[:, "Wind"].to_frame()

    aggregation_wind = tsam.TimeSeriesAggregation(
        raw_wind, noTypicalPeriods=8, hoursPerPeriod=24, clusterMethod="hierarchical"
    )

    aggregation_wind._preProcessTimeSeries()

    test = aggregation_wind.normalizedPeriodlyProfiles

    orig = pd.read_csv(
        os.path.join(
            os.path.dirname(__file__),
            "..",
            "examples",
            "results",
            "preprocessed_wind.csv",
        ),
        index_col=[0],
        header=[0, 1],
    )

    np.testing.assert_array_almost_equal(test.values, orig.values, decimal=15)


if __name__ == "__main__":
    test_preprocess()



# -*- coding: utf-8 -*-

import copy

import numpy as np

import tqdm

from tsam.timeseriesaggregation import TimeSeriesAggregation
        
def getNoPeriodsForDataReduction(noRawTimeSteps, segmentsPerPeriod, dataReduction):
    """        
    Identifies the maximum number of periods which can be set to achieve the required data reduction.

    :param noRawTimeSteps: Number of original time steps. required
    :type noRawTimeSteps: int

    :param segmentsPerPeriod: Segments per period. required
    :type segmentsPerPeriod: int

    :param dataReduction: Factor by which the resulting dataset should be reduced. required
    :type dataReduction: float

    :returns: **noTypicalPeriods** --  Number of typical periods that can be set.
    """
    return int(np.floor(dataReduction * float(noRawTimeSteps)/segmentsPerPeriod))

def getNoSegmentsForDataReduction(noRawTimeSteps, typicalPeriods, dataReduction):
    """        
    Identifies the maximum number of segments which can be set to achieve the required data reduction.

    :param noRawTimeSteps: Number of original time steps. required
    :type noRawTimeSteps: int

    :param typicalPeriods: Number of typical periods. required
    :type typicalPeriods: int

    :param dataReduction: Factor by which the resulting dataset should be reduced. required
    :type dataReduction: float

    :returns: **segmentsPerPeriod** --  Number of segments per period that can be set.
    """
    return int(np.floor(dataReduction * float(noRawTimeSteps)/typicalPeriods))




class HyperTunedAggregations(object):

    def __init__(self, base_aggregation, saveAggregationHistory=True):
        """
        A class that does a parameter variation and tuning of the aggregation itself.

        :param base_aggregation: TimeSeriesAggregation object which is used as basis for tuning the hyper parameters. required
        :type base_aggregation: TimeSeriesAggregation

        :param saveAggregationHistory: Defines if all aggregations that are created during the tuning and iterations shall be saved under self.aggregationHistory.
        :type saveAggregationHistory: boolean
        """
        self.base_aggregation = base_aggregation

        if not isinstance(self.base_aggregation, TimeSeriesAggregation):
            raise ValueError(
                "base_aggregation has to be an TimeSeriesAggregation object"
            )

        self._alterableAggregation=copy.deepcopy(self.base_aggregation)

        self.saveAggregationHistory=saveAggregationHistory

        self._segmentHistory=[]

        self._periodHistory=[]

        self._RMSEHistory=[]

        if self.saveAggregationHistory:
            self.aggregationHistory=[]
            

        
        
    def _testAggregation(self, noTypicalPeriods, noSegments):
        """
        Tests the aggregation for a set of typical periods and segments and returns the RMSE
        """
        self._segmentHistory.append(noSegments)

        self._periodHistory.append(noTypicalPeriods)

        self._alterableAggregation.noTypicalPeriods=noTypicalPeriods

        self._alterableAggregation.noSegments=noSegments

        self._alterableAggregation.createTypicalPeriods()

        self._alterableAggregation.predictOriginalData()

        RMSE=self._alterableAggregation.totalAccuracyIndicators()["RMSE"]

        self._RMSEHistory.append(RMSE)

        if self.saveAggregationHistory:
            self.aggregationHistory.append(copy.copy(self._alterableAggregation))

        return RMSE
	
    def _deleteTestHistory(self, index):
        """
        Delelets the defined index from the test history
        """
        del self._segmentHistory[index]
        del self._periodHistory[index]
        del self._RMSEHistory[index]
        
        if self.saveAggregationHistory:
            del self.aggregationHistory[index]

				
    def identifyOptimalSegmentPeriodCombination(self, dataReduction):
        """        
        Identifies the optimal combination of number of typical periods and number of segments for a given data reduction set.

        :param dataReduction: Factor by which the resulting dataset should be reduced. required
        :type dataReduction: float

        :returns: **noSegments, noTypicalperiods** --  The optimal combination of segments and typical periods for the given optimization set.
        """
        if not self.base_aggregation.segmentation:
            raise ValueError("This function does only make sense in combination with 'segmentation' activated.")

        noRawTimeSteps=len(self.base_aggregation.timeSeries.index)
        
        _maxPeriods = int(float(noRawTimeSteps)/self.base_aggregation.timeStepsPerPeriod)
        _maxSegments = self.base_aggregation.timeStepsPerPeriod

        # save RMSE
        RMSE_history = []

        # correct 0 index of python
        possibleSegments = np.arange(_maxSegments)+1
        possiblePeriods = np.arange(_maxPeriods)+1
        
        # number of time steps of all combinations of segments and periods
        combinedTimeSteps = np.outer(possibleSegments, possiblePeriods)
        # reduce to valid combinations for targeted data reduction
        reductionValidCombinations = combinedTimeSteps <= noRawTimeSteps * dataReduction

        # number of time steps for all feasible combinations
        reductionValidTimsteps = combinedTimeSteps * reductionValidCombinations
        
        # identify max segments and max period combination
        optimalPeriods = np.zeros_like(reductionValidTimsteps)
        optimalPeriods[np.arange(reductionValidTimsteps.shape[0]), reductionValidTimsteps.argmax(axis=1)] = 1
        optimalSegments = np.zeros_like(reductionValidTimsteps)
        optimalSegments[reductionValidTimsteps.argmax(axis=0), np.arange(reductionValidTimsteps.shape[1])] = 1
        
        optimalIndexCombo = np.nonzero(optimalPeriods*optimalSegments)        

        
        for segmentIx, periodIx in tqdm.tqdm(zip(optimalIndexCombo[0],optimalIndexCombo[1])):

            # derive new typical periods and derive rmse
            RMSE_history.append(self._testAggregation(possiblePeriods[periodIx], possibleSegments[segmentIx]))
                                                    
        # take the negative backwards index with the minimal RMSE 
        min_index = - list(reversed(RMSE_history)).index(min(RMSE_history)) - 1
        RMSE_min = RMSE_history[min_index]
        
                
        noTypicalPeriods=self._periodHistory[min_index]
        noSegments=self._segmentHistory[min_index]

        # and return the segment and typical period pair
        return noSegments, noTypicalPeriods, RMSE_min


    def identifyParetoOptimalAggregation(self, untilTotalTimeSteps=None):
        """        
        Identifies the pareto-optimal combination of number of typical periods and number of segments along with a steepest decent approach, starting from the aggregation to a single period and a single segment up to the representation of the full time series.
        
        :param untilTotalTimeSteps: Number of timesteps until which the pareto-front should be determined. If None, the maximum number of timesteps is chosen.
        :type untilTotalTimeSteps: int


        :returns: **** -- Nothing. Check aggregation history for results. All typical Periods in scaled form.
        """
        if not self.base_aggregation.segmentation:
            raise ValueError("This function does only make sense in combination with 'segmentation' activated.")

        noRawTimeSteps=len(self.base_aggregation.timeSeries.index)
        
        _maxPeriods = int(float(noRawTimeSteps)/self.base_aggregation.timeStepsPerPeriod)
        _maxSegments = self.base_aggregation.timeStepsPerPeriod
        
        if untilTotalTimeSteps is None:
            untilTotalTimeSteps=noRawTimeSteps


        progressBar = tqdm.tqdm(total=untilTotalTimeSteps)

        # starting point
        noTypicalPeriods=1
        noSegments=1
        _RMSE_0=self._testAggregation(noTypicalPeriods, noSegments)

        # loop until either segments or periods have reached their maximum
        while (noTypicalPeriods<_maxPeriods and noSegments<_maxSegments 
            and (noSegments+1)*noTypicalPeriods<=untilTotalTimeSteps
            and noSegments*(noTypicalPeriods+1)<=untilTotalTimeSteps):
            # test for more segments
            RMSE_segments = self._testAggregation(noTypicalPeriods, noSegments+1)
            # test for more periods
            RMSE_periods = self._testAggregation(noTypicalPeriods+1, noSegments)
            
            # RMSE old
            RMSE_old = self._RMSEHistory[-3]
            
            # segment gradient (RMSE improvement per increased time step number)
            # for segments: for each period on segment added
            RMSE_segment_gradient = (RMSE_old - RMSE_segments) / noTypicalPeriods
            # for periods: one period with no of segments
            RMSE_periods_gradient = (RMSE_old - RMSE_periods) / noSegments

            # go along the steeper gradient
            if RMSE_periods_gradient>RMSE_segment_gradient:
                noTypicalPeriods+=1
                # and delete the search direction which was not persued
                self._deleteTestHistory(-2)
            else:
                noSegments+=1
                self._deleteTestHistory(-1)
            progressBar.update(noSegments*noTypicalPeriods-progressBar.n)
            
        # afterwards loop over periods and segments exclusively until maximum is reached
        while noTypicalPeriods<_maxPeriods and noSegments*(noTypicalPeriods+1)<=untilTotalTimeSteps:
            noTypicalPeriods+=1
            RMSE = self._testAggregation(noTypicalPeriods, noSegments)
            progressBar.update(noSegments*noTypicalPeriods-progressBar.n)

        while noSegments<_maxSegments and (noSegments+1)*noTypicalPeriods<=untilTotalTimeSteps:
            noSegments+=1
            RMSE = self._testAggregation(noTypicalPeriods, noSegments)
            progressBar.update(noSegments*noTypicalPeriods-progressBar.n)
        return 



import os
import time

import pandas as pd
import numpy as np

import tsam.timeseriesaggregation as tsam


def test_weightingFactors():

    hoursPerPeriod = 24

    noTypicalPeriods = 8

    weightDict1 = {"GHI": 1, "T": 1, "Wind": 1, "Load": 1}

    weightDict2 = {"GHI": 2, "T": 2, "Wind": 2, "Load": 2}

    weightDict3 = {"GHI": 2, "T": 1, "Wind": 1, "Load": 1}

    raw = pd.read_csv(
        os.path.join(os.path.dirname(__file__), "..", "examples", "testdata.csv"),
        index_col=0,
    )

    aggregation1 = tsam.TimeSeriesAggregation(
        raw,
        noTypicalPeriods=noTypicalPeriods,
        hoursPerPeriod=hoursPerPeriod,
        clusterMethod="hierarchical",
        weightDict=weightDict1,
    )

    aggregation2 = tsam.TimeSeriesAggregation(
        raw,
        noTypicalPeriods=noTypicalPeriods,
        hoursPerPeriod=hoursPerPeriod,
        clusterMethod="hierarchical",
        weightDict=weightDict2,
    )

    aggregation3 = tsam.TimeSeriesAggregation(
        raw,
        noTypicalPeriods=noTypicalPeriods,
        hoursPerPeriod=hoursPerPeriod,
        clusterMethod="hierarchical",
        weightDict=weightDict3,
    )

    # make sure that the accuracy indicators stay the same when the different attributes are equally overweighted
    np.testing.assert_almost_equal(
        aggregation1.accuracyIndicators().values,
        aggregation2.accuracyIndicators().values,
        decimal=6,
    )

    # make sure that the RMSE of GHI is less while the other RMSEs are bigger, when GHI is overweighted
    np.testing.assert_array_less(
        aggregation3.accuracyIndicators().loc["GHI", "RMSE"],
        aggregation1.accuracyIndicators().loc["GHI", "RMSE"],
    )
    np.testing.assert_array_less(
        aggregation1.accuracyIndicators().loc[["Load", "T", "Wind"], "RMSE"],
        aggregation3.accuracyIndicators().loc[["Load", "T", "Wind"], "RMSE"],
    )


if __name__ == "__main__":
    test_weightingFactors()



import os, setuptools

dir_path = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(dir_path, "requirements.txt")) as f:
    required_packages = f.read().splitlines()
with open(os.path.join(dir_path, "README.md"), "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tsam",
    version="2.3.3",
    author="Leander Kotzur, Maximilian Hoffmann",
    author_email="leander.kotzur@googlemail.com, maximilian.hoffmann@julumni.fz-juelich.de",
    description="Time series aggregation module (tsam) to create typical periods",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/FZJ-IEK3-VSA/tsam",
    include_package_data=True,
    python_requires='>=3.9',
    packages=setuptools.find_packages(),
    install_requires=required_packages,
    setup_requires=["setuptools-git"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords=["clustering", "optimization"],
)



import tsam.timeseriesaggregation as tsam
import pandas as pd

raw = pd.read_csv("testdata.csv", index_col=0)

aggregation = tsam.TimeSeriesAggregation(
    raw, noTypicalPeriods=8, hoursPerPeriod=24, clusterMethod="hierarchical"
)
df = aggregation.createTypicalPeriods()
weights = aggregation.clusterPeriodNoOccur
aggregation.clusterOrder

timesteps = [i for i in range(0, len(df.index.get_level_values("TimeStep")))]

print(aggregation.clusterCenterIndices)

# get all index for every hour that in day of the clusterCenterIndices
days = [d for d in raw.index.dayofyear if d in aggregation.clusterCenterIndices]

# select the dates based on this
dates = raw.iloc[days].index

# TODO: Check index start (1 in dataframe, 0 in numpy -> should fit, otherwise
# days are offset one day
